
import java.awt.event.*;
import javax.swing.*;
import java.util.*;
import java.net.*;


/**
 * Base class for actions of this plug in
 *
 * @author Ulrich Hilger
 * @author Light Development
 * @author <a href="http://www.lightdev.com">http://www.lightdev.com</a>
 * @author <a href="mailto:info@lightdev.com">info@lightdev.com</a>
 *
 * @version version 1, release 1, May 1, 2003
 */

public abstract class AbstractAction2 extends javax.swing.AbstractAction implements SHTMLAction {

  protected AbstractPlugin owner;

  /**
   * constructor
   */
  public AbstractAction2(String name, AbstractPlugin owner) {
    super(name);
    this.owner = owner;
    getProperties();
  }

  /**
   * get action properties from the associated resource bundle
   *
   * @param action the action to apply properties to
   * @param cmd the name of the action to get properties for
   */
  private void getActionProperties(Action action, String cmd) {
    String path = owner.getOwner().getDynRes().getResourceString(owner.resources, cmd + DynamicResource.imageSuffix);
    if(path != null) {
      URL url = owner.getClass().getResource(path);
      if(url != null) {
        Icon icon = new ImageIcon(url);
        if (icon != null) {
          action.putValue(Action.SMALL_ICON, icon);
        }
      }
    }
    String toolTip = owner.getOwner().getDynRes().getResourceString(owner.resources, cmd + DynamicResource.toolTipSuffix);
    if(toolTip != null) {
      action.putValue(Action.SHORT_DESCRIPTION, toolTip);
    }
  }

  /**
   * update the action's state
   */
  public void update() {
    this.setEnabled(true);
  }

  /**
   * this method should be called from the constructor
   * of each AppAction and can be used to get
   * action properties from a resource file
   */
  public void getProperties() {
    getActionProperties(this, (String) getValue(Action.NAME));
  }
}