import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.text.BadLocationException;
import javax.swing.text.*;

/** This class creates the dialog box that pops up when the GoTo sub-menu of the Edit menu is selected.
 * The dialog box is used to go to a particular line in the text.<br>
 * If an invalid line number is entered, it displays an error message.
 *
 * @author TerpWord Group
 *
 * @version 3.0
 *
 * @see javax.swing.JDialog
 */
public class GoToDialog extends JDialog 
{
	private JTextField lineText;
    
	//----------------------------------------------------------
	// Constructor
	//----------------------------------------------------------
	/** Creates the dialog box that comes up when the GoTo sub-menu of the Edit menu is selected.
	 * This is used to go to a particular line in the text.
	 *
	 * @param myParentWindow The parent TwoPad window for which the GoTo command is executed.
	 * @param myContentArea The text area for which the GoTo command is executed.
	 */
	public GoToDialog(SHTMLEditorPane editor) 
	{
		setTitle("Go To Line");
		centerOnScreen();
		final SHTMLEditorPane nEditor = editor;
		getContentPane().setLayout(new BoxLayout(getContentPane(), BoxLayout.X_AXIS));
		JPanel linePanel = new JPanel();
		linePanel.setLayout(new GridLayout(2, 1, 5, 5));
		linePanel.setBorder(new EmptyBorder(5, 5, 5, 2));
		JLabel lineLabel = new JLabel("Line Number:");
		lineText = new JTextField();
		lineLabel.setLabelFor(lineText);
		linePanel.add(lineLabel);
		linePanel.add(lineText);
		getContentPane().add(linePanel);
		JPanel buttonPanel = new JPanel();
		buttonPanel.setLayout(new GridLayout(2, 1, 5, 5));
		buttonPanel.setBorder(new EmptyBorder(5, 3, 5, 5));
		JButton okButton = new JButton("OK");
		okButton.addActionListener(new ActionListener() 
		{ 
			public void actionPerformed(ActionEvent event) 
			{
				try 
				{ 
					goToLine(Integer.parseInt(lineText.getText()), nEditor); 
				}
				catch(Exception e) 
				{ 
					JOptionPane.showMessageDialog(null, "Invalid line number", "Error", JOptionPane.ERROR_MESSAGE); 
				}
				dispose();
			} 
		});
		
		buttonPanel.add(okButton);
		JButton cancelButton = new JButton("Cancel");
		cancelButton.addActionListener(new ActionListener() 
		{
			public void actionPerformed(ActionEvent event) 
			{
				dispose();
			} 
		});
		
		buttonPanel.add(cancelButton);
		getContentPane().add(buttonPanel);
		pack();
		//setLocationRelativeTo(parentWindow);
		setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
		setResizable(false);
		setVisible(true);
	}
	/**
	 * Puts the carret in the linenumber passed in.
	 * @param lineNumber the linenumber to put the cursor at.
	 * @param editor The editor where the cursor is
	 * @throws BadLocationException
	 */
	public void goToLine(int lineNumber, SHTMLEditorPane editor) throws BadLocationException 
	{
		Document doc = editor.getDocument();
		if (lineNumber == 1) {
			editor.setCaretPosition(0);
			return;
		}

		int currentLine = 1;
		for (int counter = 0; counter < doc.getLength() + 2; counter++) 
		{
			if (doc.getText(counter, 1).equals("\n")) 
			{
				currentLine++;
				if (currentLine == lineNumber) 
				{
					editor.setCaretPosition(counter + 1);
					break;
				}
			}
		}
	}
	/**
	 * centers the gotoDialog on the screen.
	 *
	 */
	public void centerOnScreen()
	{
			Dimension screen = getToolkit().getScreenSize();
			Rectangle bounds = getBounds();
			setLocation( (screen.width - bounds.width) / 2, (screen.height - bounds.height) / 2 );
			requestFocus();  
	}
}